from typing import Optional, Sequence

from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from mpl_toolkits.axes_grid1 import make_axes_locatable

import numpy as np
import anndata as ad
import scanpy as sc
import squidpy as sq

from sklearn.metrics import adjusted_rand_score

import pandas as pd
import seaborn as sns

from spicemix.model import SpiceMixPlus

def setup_squarish_axes(num_axes):
    """Create matplotlib subplots as squarely as possible."""

    height = int(np.sqrt(num_axes))
    width = num_axes // height
    height += (width * height != num_axes)
    fig, axes = plt.subplots(height, width, squeeze=False, constrained_layout=True, dpi=600, sharex=True, sharey=True)

    return fig, axes
    
def plot_metagene_embedding(trained_model: SpiceMixPlus, metagene_index: int, axes: Optional[Sequence[Axes]] = None):
    r"""Plot a single metagene in-situ across all datasets.

    Args:
        trained_model: the trained SpiceMixPlus model.
        metagene_index: the index of the metagene to plot.
        axes: A predefined set of matplotlib axes to plot on.

    """
    datasets = trained_model.datasets

    if axes is None:
        fig, axes = setup_squarish_axes(len(datasets))

    for dataset, ax in zip(datasets, axes.flat):
        dataset.plot_metagene_embedding(metagene_index, ax=ax)

    return fig

def leiden(trained_model: SpiceMixPlus, use_rep="normalized_X", joint: bool = False, resolution: float = 1.0, target_clusters: Optional[int] = None, tolerance: float = 0.05):
    r"""Compute Leiden clustering for all datasets.

    Args:
        trained_model: the trained SpiceMixPlus model.
        joint: if `True`, jointly cluster the spots
        use_rep: the key in the ``.obsm`` dataframe to ue as input to the Leiden clustering algorithm.
        resolution: the resolution to use for Leiden clustering. Higher values yield finer clusters..
    """
    # TODO: implement joint clustering
   
    cluster(trained_model, use_rep=use_rep, joint=joint, resolution=resolution, target_clusters=target_clusters, tolerance=tolerance)

def cluster(trained_model: SpiceMixPlus, use_rep="normalized_X", joint: bool = False, method: str = "leiden", resolution: float = 1.0, target_clusters: Optional[int] = None, tolerance: float = 0.05):
    r"""Compute clustering for all datasets.

    Args:
        trained_model: the trained SpiceMixPlus model.
        joint: if `True`, jointly cluster the spots
        use_rep: the key in the ``.obsm`` dataframe to ue as input to the Leiden clustering algorithm.
        resolution: the resolution to use for Leiden clustering. Higher values yield finer clusters..
    """
    # TODO: implement joint clustering
    
    datasets = trained_model.datasets
    if joint:
        dataset_names = [dataset.name for dataset in datasets]
        merged_dataset = ad.concat(datasets, label="batch", keys=dataset_names, merge="unique", uns_merge="unique", pairwise=True)
        datasets = [merged_dataset]

        
    clustering_function = getattr(sc.tl, method)
    for dataset in datasets:
        sc.pp.neighbors(dataset, use_rep=use_rep)
        clustering_function(dataset, resolution=resolution)
    
        num_clusters = len(dataset.obs[method].unique())
        print(num_clusters)
       
        lower_bound = 0.25 * resolution 
        upper_bound = 1.75 * resolution 
        while target_clusters and num_clusters != target_clusters and np.abs(lower_bound - upper_bound) > tolerance:
            effective_resolution = (lower_bound * upper_bound) ** 0.5
            clustering_function(dataset, resolution=effective_resolution)
            num_clusters = len(dataset.obs[method].unique())
            if num_clusters < target_clusters:
                lower_bound = effective_resolution
            elif num_clusters >= target_clusters:
                upper_bound = effective_resolution
            print(num_clusters)

    if joint:
        indices = merged_dataset.obs.groupby("batch").indices.values()
        unmerged_datasets = [merged_dataset[index] for index in indices]
        for unmerged_dataset, original_dataset in zip(unmerged_datasets, trained_model.datasets):
            original_dataset.obs[method] = unmerged_dataset.obs[method]


def plot_in_situ(trained_model: SpiceMixPlus, color="leiden", axes = None, **spatial_kwargs):
    r"""Plot a categorical label across all datasets in-situ.

    Extends AnnData's ``sc.pl.spatial`` function to plot labels/values across multiple replicates.

    Args:
        trained_model: the trained SpiceMixPlus model.
        color: the key in the ``.obs`` dataframe to plot.
        axes: A predefined set of matplotlib axes to plot on.
    """
    datasets = trained_model.datasets
        
    fig = None
    if axes is None:
        fig, axes = setup_squarish_axes(len(datasets))

    edges_width = 0.2 if "edges_width" not in spatial_kwargs else spatial_kwargs.pop("edges_width")
    spot_size = 0.04 if "spot_size" not in spatial_kwargs else spatial_kwargs.pop("spot_size")
    edges = True if "edges" not in spatial_kwargs else spatial_kwargs.pop("edges")
    palette = sc.pl.palettes.godsnot_102 if "palette" not in spatial_kwargs else spatial_kwargs.pop("palette")
    legend_fontsize = "xx-small" if "legend_fontsize" not in spatial_kwargs else spatial_kwargs.pop("legend_fontsize")

    neighbors_key = "spatial_neighbors" if "spatial_neighbors" not in spatial_kwargs else spatial_kwargs.pop("neighbors_key")
    for dataset, ax in zip(datasets, axes.flat):
        sc.pl.spatial(dataset, spot_size=spot_size, neighbors_key=neighbors_key,
            color=color, edges=edges,  edges_width=edges_width, legend_fontsize=legend_fontsize,
            ax=ax, show=False, palette=palette, **spatial_kwargs)

def plot_umap(trained_model: SpiceMixPlus, color="leiden", axes = None, **_kwargs):
    r"""Plot a categorical label across all datasets in-situ.

    Extends AnnData's ``sc.pl.spatial`` function to plot labels/values across multiple replicates.

    Args:
        trained_model: the trained SpiceMixPlus model.
        color: the key in the ``.obs`` dataframe to plot.
        axes: A predefined set of matplotlib axes to plot on.
    """
    datasets = trained_model.datasets
        
    fig = None
    if axes is None:
        fig, axes = setup_squarish_axes(len(datasets))

    edges_width = 0.2 if "edges_width" not in spatial_kwargs else spatial_kwargs.pop("edges_width")
    spot_size = 0.04 if "spot_size" not in spatial_kwargs else spatial_kwargs.pop("spot_size")
    edges = True if "edges" not in spatial_kwargs else spatial_kwargs.pop("edges")
    palette = sc.pl.palettes.godsnot_102 if "palette" not in spatial_kwargs else spatial_kwargs.pop("palette")
    legend_fontsize = "xx-small" if "legend_fontsize" not in spatial_kwargs else spatial_kwargs.pop("legend_fontsize")
    for dataset, ax in zip(datasets, axes.flat):
        sc.pl.spatial(dataset, spot_size=spot_size, neighbors_key="spatial_neighbors",
            color=color, edges=True,  edges_width=edges_width, legend_fontsize=legend_fontsize,
            ax=ax, show=False, palette=palette, **spatial_kwargs)

def multireplicate_heatmap(trained_model: SpiceMixPlus,
    title_font_size: Optional[int] = None,
    axes: Optional[Sequence[Axes]] = None,
    obsm: Optional[str] = None,
    obsp: Optional[str] = None,
    uns: Optional[str] = None,
    **heatmap_kwargs
  ):
    r"""Plot 2D heatmap data across all datasets.

    Wrapper function to enable plotting of continuous 2D data across multiple replicates. Only
    one of ``obsm``, ``obsp`` or ``uns`` should be used.

    Args:
        trained_model: the trained SpiceMixPlus model.
        axes: A predefined set of matplotlib axes to plot on.
        obsm: the key in the ``.obsm`` dataframe to plot.
        obsp: the key in the ``.obsp`` dataframe to plot.
        uns: the key in the ``.uns`` dataframe to plot. Unstructured data must be 2D in shape.
        **heatmap_kwargs: arguments to pass to the `ax.imshow` call for each dataset
    """
    datasets = trained_model.datasets
    

    fig = None
    if axes is None:
        fig, axes = setup_squarish_axes(len(datasets))

    aspect = 0.05 if "aspect" not in heatmap_kwargs else heatmap_kwargs.pop("aspect")    
    cmap = "hot" if "cmap" not in heatmap_kwargs else heatmap_kwargs.pop("cmap")    

    for dataset_index, ax in enumerate(axes.flat):
        if dataset_index >= len(datasets):
            ax.set_visible(False)
            continue
        
        dataset = datasets[dataset_index]
        key = None
        if obsm:
            image = dataset.obsm[obsm][dataset.name]
        if obsp:
            image = dataset.obsp[obsp][dataset.name]
        if uns:
            image = dataset.uns[uns][dataset.name]
       
        im = ax.imshow(image, cmap=cmap, interpolation='nearest', aspect=aspect, **heatmap_kwargs)
        if title_font_size is not None:
            ax.set_title(dataset.name, fontsize= title_font_size)

        fig.colorbar(im, ax=ax, orientation='vertical')


def multigroup_heatmap(trained_model: SpiceMixPlus,
    title_font_size: Optional[int] = None,
    group_type: str = "metagene",
    axes: Optional[Sequence[Axes]] = None,
    key: Optional[str] = None,
    **heatmap_kwargs
  ):
    r"""Plot 2D heatmap data across all datasets.

    Wrapper function to enable plotting of continuous 2D data across multiple replicates. Only
    one of ``obsm``, ``obsp`` or ``uns`` should be used.

    Args:
        trained_model: the trained SpiceMixPlus model.
        axes: A predefined set of matplotlib axes to plot on.
        obsm: the key in the ``.obsm`` dataframe to plot.
        obsp: the key in the ``.obsp`` dataframe to plot.
        uns: the key in the ``.uns`` dataframe to plot. Unstructured data must be 2D in shape.
        **heatmap_kwargs: arguments to pass to the `ax.imshow` call for each dataset
    """
    datasets = trained_model.datasets
    groups = trained_model.metagene_groups if group_type == "metagene" else trained_model.spatial_affinity_groups
    

    fig = None
    if axes is None:
        fig, axes = setup_squarish_axes(len(groups))

    aspect = 0.05 if "aspect" not in heatmap_kwargs else heatmap_kwargs.pop("aspect")    
    cmap = "hot" if "cmap" not in heatmap_kwargs else heatmap_kwargs.pop("cmap")    

    for group_index, (ax, group_name) in enumerate(zip(axes.flat, groups)):
        first_dataset_name = groups[group_name][0]
        first_dataset = next(filter(lambda dataset: dataset.name == first_dataset_name, datasets))

        if group_index > len(groups):
            ax.set_visible(False)
            continue
        
        image = first_dataset.uns[key][group_name]
       
        im = ax.imshow(image, cmap=cmap, interpolation='nearest', aspect=aspect, **heatmap_kwargs)
        if title_font_size is not None:
            ax.set_title(group_name, fontsize= title_font_size)
        fig.colorbar(im, ax=ax, orientation='vertical')

def compute_ari_scores(trained_model: SpiceMixPlus, labels: str, predictions: str, ari_key: str = "ari"):
    r"""Compute adjusted Rand index (ARI) score  between a set of ground truth labels and an unsupervised clustering.

    Useful for assessing clustering validity. ARI score is computed per dataset.

    Args:
        trained_model: the trained SpiceMixPlus model.
        labels: the key in the ``.obs`` dataframe for the label data.
        predictions: the key in the ``.obs`` dataframe for the predictions data.
        ari_key: the key in the ``.uns`` dictionary where the ARI score will be stored.
    """
    datasets = trained_model.datasets

    for dataset in datasets:
        ari = adjusted_rand_score(dataset.obs[labels], dataset.obs[predictions])
        dataset.uns[ari_key] = ari

def plot_all_metagene_embeddings(trained_model: SpiceMixPlus, embedding_key: str = "X", column_names: Optional[str] = None, **spatial_kwargs):
    r"""Plot all laerned metagenes in-situ across all replicates.

    Each replicate's metagenes are contained in a separate plot.

    Args:
        trained_model: the trained SpiceMixPlus model.
        embedding_key: the key in the ``.obsm`` dataframe for the cell/spot embeddings.
        column_names: a list of the suffixes for each latent feature. If ``None``, it is assumed
            that these suffixes are just the indices of the latent features.
    """

    if column_names == None:
        column_names = [f"{embedding_key}_{index}" for index in range(trained_model.K)]

    datasets = trained_model.datasets
    spot_size = 0.1 if "spot_size" not in spatial_kwargs else spatial_kwargs.pop("spot_size")
    palette = sc.pl.palettes.godsnot_102 if "palette" not in spatial_kwargs else spatial_kwargs.pop("palette")
    for dataset in datasets:
        axes = sc.pl.spatial(
            sq.pl.extract(dataset, embedding_key, prefix=f"{embedding_key}"),
            color=column_names,
            spot_size=spot_size,
            wspace=0.2,
            ncols=2,
        )

def compute_empirical_correlations(trained_model: SpiceMixPlus, feature: str = "X", output: str = "empirical_correlation"):
    """Compute the empirical spatial correlation for a feature set across all datasets.

    Args:
        trained_model: the trained SpiceMixPlus model.
        feature: key in `.obsm` of feature set for which spatial correlation should be computed.
        output: key in `.uns` where output correlation matrices should be stored.
    """

    datasets = trained_model.datasets
    num_replicates = len(datasets)
    K = trained_model.K
    scaling = trained_model.parameter_optimizer.spatial_affinity_state.scaling
    empirical_correlations = np.zeros([num_replicates, K, K])
    for replicate, dataset in enumerate(datasets):
        adjacency_list = dataset.obs["adjacency_list"]
        X = dataset.obsm[feature]
        Z = X / np.linalg.norm(X, axis=1, keepdims=True, ord=1)
        edges = np.array([(i, j) for i, e in enumerate(adjacency_list) for j in e])

        x = Z[edges[:, 0]]
        y = Z[edges[:, 1]]
        x = x - x.mean(axis=0, keepdims=True)
        y = y - y.mean(axis=0, keepdims=True)
        y_std = y.std(axis=0, keepdims=True)
        x_std = x.std(axis=0, keepdims=True)
        corr = (y / y_std).T @ (x / x_std) / len(x)
        empirical_correlations[replicate] = - corr

    # Symmetrizing and zero-centering empirical_correlation
    empirical_correlations = (empirical_correlations + np.transpose(empirical_correlations, (0, 2, 1))) / 2
    empirical_correlations -= empirical_correlations.mean(axis=(1, 2), keepdims=True)
    empirical_correlations *= scaling

    for dataset, empirical_correlation in zip(datasets, empirical_correlations):
        all_correlations = {dataset.name: empirical_correlation}
        dataset.uns[output] = all_correlations




import sys, logging, time, gc, os
from tqdm.auto import tqdm, trange

import numpy as np
import torch

from sample_for_integral import project2simplex
from util import NesterovGD

@torch.no_grad()
def estimate_weight_wonbr(Y, M, X, sigma_yx, replicate, prior_x_mode, prior_x, dataset, context, n_epochs=1000, tol=1e-6, update_alg='gd', verbose=True):
    """Estimate weights without spatial information - equivalent to vanilla NMF.

    min 1/2σ^2 || Y - X MT ||_2^2 + lam || X ||_1
    grad = X MT M / σ^2 - Y MT / σ^2 + lam

    TODO: use (projected) Nesterov GD. not urgent
    """

    M = dataset.uns["M"][f"{replicate}"]
    sigma_yx = dataset.uns["sigma_yx"]

    # Precomputing quantities 
    MTM = M.T @ M / (sigma_yx ** 2)
    YM = Y.to(M.device) @ M / (sigma_yx ** 2)
    Ynorm = torch.linalg.norm(Y, ord='fro').item() ** 2 / (sigma_yx ** 2)
    step_size = 1 / torch.linalg.eigvalsh(MTM).max().item()
    loss_prev, loss = np.inf, np.nan

    def multiplicative_update(X_prev):
        """
        TODO:UNTESTED

        """
        X = torch.clip(X_prev, min=1e-10)
        loss = ((X @ MTM) * X).sum() / 2 - clipped_X.view(-1) @ YM.view(-1) + Ynorm / 2
        numerator = YM
        denominator = X @ MTM
        if prior_x_mode == 'exponential shared fixed':
            # see sklearn.decomposition.NMF
            loss += (X @ prior_x[0]).sum()
            denominator += prior_x[0][None]
        else:
            raise NotImplementedError

        loss = loss.item()
        assert loss <= loss_prev * (1 + 1e-4), (loss_prev, loss, (loss_prev - loss) / loss)
        multiplicative_factor = numerator / denominator
        X *= multiplicative_factor
        torch.clip(X, min=1e-10)

        return X, loss

    def gradient_update(X):
        """
        TODO:UNTESTED

        """
        quadratic_term_gradient = X @ MTM
        linear_term_gradient = YM
        if prior_x_mode == 'exponential shared fixed':
            linear_term_gradient = linear_term_gradient - prior_x[0][None]
        elif not prior_x_mode:
            pass
        else:
            raise NotImplementedError
        loss = (quadratic_term_gradient * X).sum().item() / 2 - (linear_term_gradient * X).sum().item() + Ynorm / 2
        gradient = quadratic_term_gradient - linear_term_gradient
        X = X.sub(gradient, alpha=step_size)
        X = torch.clip(X, min=1e-10)
        
        return X, loss
        
    progress_bar = trange(n_epochs, leave=True, disable=not verbose, miniters=1000)
    for epoch in progress_bar:
        X_prev = X.clone()
        if update_alg == 'mu':
            # TODO: it seems like loss might not always decrease...
            X.clip_(min=1e-10)
            loss = ((X @ MTM) * X).sum() / 2 - X.view(-1) @ YM.view(-1) + Ynorm / 2
            numerator = YM
            denominator = X @ MTM
            if prior_x_mode == 'exponential shared fixed':
                # see sklearn.decomposition.NMF
                loss += (X @ prior_x[0]).sum()
                denominator.add_(prior_x[0][None])
            elif not prior_x_mode:
                pass
            else:
                raise NotImplementedError

            loss = loss.item()
            assert loss <= loss_prev * (1 + 1e-4), (loss_prev, loss, (loss_prev - loss) / loss)
            multiplicative_factor = numerator / denominator
            X.mul_(multiplicative_factor).clip_(min=1e-10)

            # X, loss = multiplicative_update(X_prev)
        elif update_alg == 'gd':
            X, loss = gradient_update(X)
        else:
            raise NotImplementedError

        dX = torch.abs((X_prev - X) / torch.linalg.norm(X, dim=1, ord=1, keepdim=True)).max().item()
        do_stop = dX < tol
        if epoch % 1000 == 0 or do_stop:
            progress_bar.set_description(
                f'Updating weight w/o nbrs: loss = {loss:.1e} '
                f'%δloss = {(loss_prev - loss) / loss:.1e} '
                f'%δX = {dX:.1e}'
            )
        loss_prev = loss
        if do_stop:
            break
    progress_bar.close()
    return loss, X


class IndependentSet:
    """Iterator class that yields a list of batch_size independent nodes from a spatial graph.

    For each iteration, no pair of yielded nodes can be neighbors of each other according to the
    adjacency matrix.

    """

    def __init__(self, adjacency_list, batch_size=50):
        self.N = len(adjacency_list)
        self.adjacency_list = adjacency_list
        self.batch_size = batch_size
        self.indices_remaining = None

    def __iter__(self):
        self.indices_remaining = set(range(self.N))
        return self

    def __next__(self):
        # make sure selected nodes are not adjacent to each other
        # i.e., find an independent set of `indices_candidates` in a greedy manner
        if len(self.indices_remaining) == 0:
            raise StopIteration

        valid_indices = []
        excluded_indices = set()
        effective_batch_size = min(self.batch_size, len(self.indices_remaining))
        candidate_indices = np.random.choice(list(self.indices_remaining),
            size=effective_batch_size,
            replace=False,
        )
        for index in candidate_indices:
            if index not in excluded_indices:
                valid_indices.append(index)
                excluded_indices |= set(self.adjacency_list[index])

        self.indices_remaining -= set(valid_indices)

        return valid_indices

@torch.no_grad()
def estimate_weight_wnbr(Y, M, X, sigma_yx, replicate, prior_x_mode, prior_x, dataset, context, n_epochs=1000, tol=1e-5, update_alg='nesterov'):
    """Estimate updated weights taking neighbor-neighbor interactions into account.

    The optimization for all variables
    min 1/2σ^2 || Y - diag(S) Z MT ||_2^2 + lam || S ||_1 + sum_{ij in E} ziT Σx-1 zj

    for s_i
    min 1/2σ^2 || y - M z s ||_2^2 + lam s
    s* = max(0, ( yT M z / σ^2 - lam ) / ( zT MT M z / σ^2) )

    for Z
    min 1/2σ^2 || Y - diag(S) Z MT ||_2^2 + sum_{ij in E} ziT Σx-1 zj
    grad_i = MT M z s^2 / σ^2 - MT y s / σ^2 + sum_{j in Ei} Σx-1 zj

    TODO: Try projected Newton's method.
    TM: Inverse is precomputed once, and projection is cheap. Not sure if it works theoretically
    """
    M = dataset.uns["M"][f"{replicate}"]

    # Precomputing quantities
    MTM = M.T @ M / (sigma_yx ** 2)
    YM = Y.to(M.device) @ M / (sigma_yx ** 2)
    Ynorm = torch.linalg.norm(Y, ord='fro').item() ** 2 / (sigma_yx ** 2)
    base_step_size = 1 / torch.linalg.eigvalsh(MTM).max().item()
    S = torch.linalg.norm(X, dim=1, ord=1, keepdim=True)
    Z = X / S
    N = len(Z)
    
    E_adjacency_list = dataset.obs["adjacency_list"]
    Sigma_x_inv = dataset.uns["Sigma_x_inv"][f"{replicate}"]

    def get_adjacency_matrix(adjacency_list):
        edges = [(i, j) for i, e in enumerate(adjacency_list) for j in e]
        adjacency_matrix = torch.sparse_coo_tensor(np.array(edges).T, np.ones(len(edges)), size=[len(adjacency_list), N], **context)
        return adjacency_matrix

    def update_s():
        # print(f"Minimum S: {S.min()}")
        S[:] = (YM * Z).sum(axis=1, keepdim=True)
        # print(f"Minimum S: {S.min()}")
        # print(f"prior_x_mode: {prior_x_mode}")
        # print(f"prior_x: {prior_x[0][0]}")
        if prior_x_mode == 'exponential shared fixed':
            # TODO: why divide by two?
            S.sub_(prior_x[0][0] / 2)
        elif not prior_x_mode:
            pass
        else:
            raise NotImplementedError

        denominator = ((Z @ MTM) * Z).sum(axis=1, keepdim=True)
        S.div_(denominator)
        S.clip_(min=1e-5)

    # def update_z_mu(Z):
    #     indices = np.random.permutation(len(S))
    #     # bs = len(indices) // 50
    #     bs = 1
    #     thr = 1e3
    #     for i in range(0, len(indices), bs):
    #         idx = indices[i: i+bs]
    #         adjacency_matrix = get_adjacency_matrix(E_adjacency_list[idx])
    #         numerator = YM[idx] * S[idx]
    #         denominator = (Z[idx] @ MTM).mul_(S[idx] ** 2)
    #         t = (adjacency_matrix @ Z) @ Sigma_x_inv
    #         # TM: not sure if projected mu works
    #         # TM: seems not
    #         numerator -= t.clip(max=0)
    #         denominator += t.clip(min=0)
    #         mul_fac = numerator / denominator
    #         mul_fac.clip_(min=1/thr, max=thr)
    #         # Z.mul_(mul_fac)
    #         Z[idx] *= mul_fac
    #         result = project2simplex(Z, dim=1)
    #         Z = 0
    #         Z = result

    #     return Z
    
    # Debugging for loss of M
    def compute_loss_and_gradient(M):
        quadratic_factor_grad = M @ quadratic_factor
        loss = (quadratic_factor_grad * M).sum()
        grad = quadratic_factor_grad
        linear_term_grad = linear_term
        loss -= 2 * (linear_term_grad * M).sum()
        grad -= linear_term_grad
        
        loss += constant
        loss /= 2

        if M_constraint == 'simplex':
            grad.sub_(grad.sum(0, keepdim=True))

        return loss.item(), grad
    
    def calc_func_grad(Z_batch, S_batch, quad, linear):
        t = (Z_batch @ quad).mul_(S_batch ** 2)
        f = (t * Z_batch).sum() / 2
        g = t
        t = linear
        f -= (t * Z_batch).sum()
        g -= t
        g.sub_(g.sum(1, keepdim=True))
        
        return f.item(), g

    def update_z_gd(Z):
        # def calc_func_grad(Z, idx, adjacency_matrix=None):
        #   # grad = (Z @ MTM).mul_(S ** 2)
        #   # grad.addcmul_(YM, S, value=-1)
        #   # grad.addmm_(E @ Z, Sigma_x_inv)
        #   # grad.sub_(grad.sum(1, keepdim=True))
        step_size = base_step_size / S.square()
        pbar = tqdm(range(N), leave=False, disable=True)
        for idx in IndependentSet(E_adjacency_list, batch_size=128):
            step_size_scale = 1
            quad_batch = MTM
            linear_batch = YM[idx] * S[idx] - get_adjacency_matrix(E_adjacency_list[idx]) @ Z @ Sigma_x_inv
            Z_batch = Z[idx].contiguous()
            S_batch = S[idx].contiguous()
            step_size_batch = step_size[idx].contiguous()
            func, grad = calc_func_grad(Z_batch, S_batch, quad_batch, linear_batch)
            while True:
                Z_batch_new = Z_batch - step_size_batch * step_size_scale * grad
                result = project2simplex(Z_batch_new, dim=1)
                Z_batch_new = 0
                Z_batch_new = result
                dZ = Z_batch_new.sub(Z_batch).abs().max().item()
                func_new, grad_new = calc_func_grad(Z_batch_new, S_batch, quad_batch, linear_batch)
                if func_new < func:
                    Z_batch = Z_batch_new
                    func = func_new
                    grad = grad_new
                    step_size_scale *= 1.1
                    continue
                else:
                    step_size_scale *= .5
                if dZ < tol or step_size_scale < .5: break
            assert step_size_scale > .1
            Z[idx] = Z_batch
            pbar.set_description(f'Updating Z w/ nbrs via line search: lr={step_size_scale:.1e}')
            pbar.update(len(idx))
        pbar.close()

        return Z

    def update_z_gd_nesterov(Z, verbose=False):
        pbar = trange(N, leave=False, disable=True, desc='Updating Z w/ nbrs via Nesterov GD')
        batch_number = 0
        func, grad = calc_func_grad(Z, S, MTM, YM * S - get_adjacency_matrix(E_adjacency_list) @ Z @ Sigma_x_inv / 2)
        # print(f"Z initial loss: {func}")
        for idx in IndependentSet(E_adjacency_list, batch_size=256):
        # for idx in [[N-1]]:
            quad_batch = MTM
            # linear_batch = YM[idx] * S[idx] - get_adjacency_matrix(E_adjacency_list[idx]) @ Z @ Sigma_x_inv
            linear_batch_spatial = - get_adjacency_matrix(E_adjacency_list[idx]) @ Z @ Sigma_x_inv
            Z_batch = Z[idx].contiguous()
            S_batch = S[idx].contiguous()
                

            # print(f"Updated step size: {base_step_size / S_batch.square()}")
            optimizer = NesterovGD(Z_batch, base_step_size / S_batch.square())
            ppbar = trange(10000, leave=False, disable=True)
            # while True:
            for i_iter in ppbar:
                # loss_old, _ = compute_loss(np.inf)
                # S_old = S.clone()
                update_s() # TODO: update S_batch directly
                # loss_new, dloss = compute_loss(loss_old)
                # print(loss_old, loss_new, dloss, S_old.sub(S).abs().max().item())
                S_batch = S[idx].contiguous()
                linear_batch = linear_batch_spatial + YM[idx] * S_batch
                if i_iter == 0:
                    func, grad = calc_func_grad(Z_batch, S_batch, quad_batch, linear_batch)
                    # print(f"Z_batch initial loss: {func}")
                    func, grad = calc_func_grad(Z, S, MTM, YM * S - get_adjacency_matrix(E_adjacency_list) @ Z @ Sigma_x_inv / 2)
                    # print(f"Z loss: {func}")
                NesterovGD.step_size = base_step_size / S_batch.square() # TM: I think this converges as s converges
                func, grad = calc_func_grad(Z_batch, S_batch, quad_batch, linear_batch)
                Z_batch_prev = Z_batch.clone()
                # loss_old, _ = compute_loss(np.inf)
                Z_batch_copy = optimizer.step(grad)
                # print(Z_batch_copy.max())
                Z_batch = project2simplex(Z_batch_copy, dim=1)
                # print(result.sum(axis=1))
                optimizer.set_parameters(Z_batch)
                dZ = (Z_batch_prev - Z_batch).abs().max().item()
                Z[idx] = Z_batch
                # loss_new, dloss = compute_loss(loss_old)
                # if i_iter % 100 == 0:
                #   print(loss_old, loss_new, dloss, dZ)
                #   d = S_batch * Z_batch - X_bak[None]
                #   print(torch.linalg.norm(d))
                #   print(d)
                ppbar.set_description(f'func={func:.1e}, dZ={dZ:.1e}')
                if dZ < tol:
                    break
            ppbar.close()
            Z[idx] = Z_batch
            func, grad = calc_func_grad(Z_batch, S_batch, quad_batch, linear_batch)
            if verbose:
                print(f"Z_batch final loss: {func}")
            func, grad = calc_func_grad(Z, S, MTM, YM * S - get_adjacency_matrix(E_adjacency_list) @ Z @ Sigma_x_inv /2 )
            if verbose:
                print(f"Z loss: {func}")
            # assert False
            # pbar.set_description(f'Updating Z w/ nbrs via Nesterov GD')
            pbar.update(len(idx))
        pbar.close()
        func, grad = calc_func_grad(Z, S, MTM, YM * S - get_adjacency_matrix(E_adjacency_list) @ Z @ Sigma_x_inv / 2)
        if verbose:
            print(f"Z final loss: {func}")

        return Z

    def compute_loss():
        X = Z * S
        loss = ((X @ MTM) * X).sum() / 2 - (X * YM).sum() + Ynorm / 2
        if prior_x_mode == 'exponential shared fixed':
            loss += prior_x[0][0] * S.sum()
        elif not prior_x_mode:
            pass
        else:
            raise NotImplementedError

        if Sigma_x_inv is not None:
            loss += ((dataset.obsp["adjacency_matrix"] @ Z) @ Sigma_x_inv).mul(Z).sum() / 2
        loss = loss.item()
        # assert loss <= loss_prev, (loss_prev, loss)
        return loss

    # TM: consider combine compute_loss and update_z to remove a call to torch.sparse.mm
    # TM: the above idea is not practical if we update only a subset of nodes each time

    loss = np.inf
    pbar = trange(n_epochs, desc='Updating weight w/ neighbors')

    for epoch in pbar:
        update_s()
        Z_prev = Z.clone().detach()
        # We may use Nesterov first and then vanilla GD in later iterations
        # update_z_mu(Z)
        # update_z_gd(Z)
        if update_alg == "gd":
            Z = update_z_gd(Z)
        elif update_alg == "nesterov":
            Z = update_z_gd_nesterov(Z)

        loss_prev = loss
        loss = compute_loss()
        dloss = loss_prev - loss
        dZ = (Z_prev - Z).abs().max().item()
        pbar.set_description(
            f'Updating weight w/ neighbors: loss = {loss:.1e} '
            f'δloss = {dloss:.1e} '
            f'δZ = {dZ:.1e}'
        )
        if dZ < tol: break

    X_final = Z * S
    return loss, X_final




import sys, os, time
from multiprocessing import Pool

import numpy as np
import torch
import scipy

def project2simplex(y, dim=0, zero_threshold=1e-10):
    """
    # https://math.stackexchange.com/questions/2402504/orthogonal-projection-onto-the-unit-simplex
    find a scalar mu such that || (y-mu)_+ ||_1 = 1

    Currently uses Newton's method to optimize || y - mu ||^2

    TODO: try implementing it this way instead: https://arxiv.org/pdf/1101.6081.pdf

    Args:
        y: vector to be projected to unit simplex
    """
    
    num_components = y.shape[dim]
    
    mu = (y.sum(dim=dim, keepdim=True) - 1) / num_components
    previous_derivative = derivative = None
    for _ in range(num_components):
        difference = y - mu
        derivative = -(difference > zero_threshold).sum(dim=dim, keepdim=True).to(y.dtype)
        assert -derivative.min() > 0, difference.clip(min=0).sum(dim).min()
        if previous_derivative is not None and (derivative == previous_derivative).all():
            break
        objective_value = torch.clip(difference, min=zero_threshold).sum(dim=dim, keepdim=True) - 1
        newton_update = objective_value / derivative
        mu -= newton_update
        previous_derivative = derivative
    assert (derivative == previous_derivative).all()
    
    assert not torch.isnan(y).any(), y
    #print(mu)
    if torch.isnan(mu).any():
        np.save("../problematic_y.npy", y.detach().cpu().numpy())

    y = (y - mu).clip(min=zero_threshold)
    assert not torch.isnan(y).any(), (mu, derivative)
    # print(y.sum(dim=dim).sub_(1))

    assert y.sum(dim=dim).sub_(1).abs_().max() < 1e-3, y.sum(dim=dim).sub_(1).abs_().max()
    
    return y


def test_project2simplex():
    x = torch.rand(100, 100)
    x = x * 3 - 1
    project2simplex(x, dim=0)

def test_project2simplex_basic():
    x = torch.tensor([1/2, 1/4, 1/4])
    assert x.allclose(project2simplex(x.clone(), dim=0))

def integrate_of_exponential_over_simplex(eta, eps=1e-15):
	assert torch.isfinite(eta).all()
	N, K = eta.shape
	A = torch.empty_like(eta)
	signs = torch.empty_like(A)
	for k in range(K):
		t = eta - eta[:, [k]]
		assert torch.isfinite(t).all()
		t[:, k] = 1
		tsign = t.sign()
		signs[:, k] = tsign.prod(-1)
		# t = t.abs().clip(min=1e-10).log()
		t = t.abs().add(eps).log()
		assert torch.isfinite(t).all()
		t[:, k] = -eta[:, k]
		A[:, k] = t.sum(-1).neg()
	assert torch.isfinite(A).all()
	# signed logsumexp
	o = A.max(-1, keepdim=True)[0]
	ret = A.sub(o).exp()
	assert torch.isfinite(ret).all()
	ret = ret.mul(signs).sum(-1)
	ret = ret.clip(min=eps)
	assert (ret > 0).all(), ret.min().item()
	ret = ret.log().add(o.squeeze(-1))
	return ret

if __name__ == "__main__":
    test_project2simplex_basic()



import sys, time, itertools, logging
from multiprocessing import Pool, Process
from tqdm.auto import tqdm, trange
from collections import defaultdict

import torch
import numpy as np
from scipy.special import loggamma

from sample_for_integral import integrate_of_exponential_over_simplex, project2simplex
from util import NesterovGD

@torch.no_grad()
def project_M(M, M_constraint):
    result = M.clone()
    if M_constraint == 'simplex':
        result = project2simplex(result, dim=0, zero_threshold=1e-5)
    elif M_constraint == 'unit sphere':
        result = M.div(torch.linalg.norm(result, ord=2, dim=0, keepdim=True))
    elif M_constraint == 'nonneg unit sphere':
        result = M.clip(1e-10).div(torch.linalg.norm(result, ord=2, dim=0, keepdim=True))
    else:
        raise NotImplementedError
    return result

@torch.no_grad()
def estimate_M(Xs, M, betas, datasets, M_constraint, context,
        M_bar=None, lambda_M=0,
        n_epochs=10000, tol=1e-6, backend_algorithm='gd Nesterov', verbose=False):
    """Optimize metagene parameters.

    M is shared across all replicates.
    min || Y - X MT ||_2^2 / (2 σ_yx^2)
    s.t. || Mk ||_p = 1
    grad = (M XT X - YT X) / (σ_yx^2)

    Each replicate may have a slightly different M
    min || Y - X MT ||_2^2 / (2 σ_yx^2) + || M - M_bar ||_2^2 λ_M / 2
    s.t. || Mk ||_p = 1
    grad = ( M XT X - YT X ) / ( σ_yx^2 ) + λ_M ( M - M_bar )

    Args:
        Ys: list of gene expression data
        Xs: list of estimated hidden states
        M (torch.Tensor): current estimate of metagene parameters
        betas: weight of each FOV in optimization scheme
        context: context ith which to create PyTorch tensor
        n_epochs: number of epochs 

    Returns:
        Updated estimate of metagene parameters.
    """

    _, K = M.shape
    quadratic_factor = torch.zeros([K, K], **context)
    linear_term = torch.zeros_like(M)
    sigma_yxs = np.array([dataset.uns["sigma_yx"] for dataset in datasets])
    scaled_betas = betas / (sigma_yxs**2)
    
    # ||Y||_2^2
    constant_magnitude = np.array([torch.linalg.norm(torch.tensor(dataset.X, **context)).item()**2 for dataset in datasets]).sum()

    constant = (np.array([torch.linalg.norm(torch.tensor(dataset.X, **context)).item()**2 for dataset in datasets]) * scaled_betas).sum()
    regularization = [dataset.uns["spicemixplus_hyperparameters"]["prior_x"] for dataset in datasets]
    for dataset, X, sigma_yx, scaled_beta in zip(datasets, Xs, sigma_yxs, scaled_betas):
        # X_c^TX_c
        quadratic_factor.addmm_(X.T, X, alpha=scaled_beta)
        # MX_c^TY_c
        linear_term.addmm_(torch.tensor(dataset.X, **context).T.to(X.device), X, alpha=scaled_beta)

    if lambda_M > 0 and M_bar is not None:
        quadratic_factor.diagonal().add_(lambda_M)
        linear_term += lambda_M * M_bar

    loss_prev, loss = np.inf, np.nan
    progress_bar = trange(n_epochs, leave=True, disable=not verbose, desc='Updating M', miniters=1000)

    def compute_loss_and_gradient(M, verbose=False):
        quadratic_factor_grad = M @ quadratic_factor
        loss = (quadratic_factor_grad * M).sum()
        if verbose:
            print(f"M quadratic term: {loss}")
        grad = quadratic_factor_grad
        linear_term_grad = linear_term
        loss -= 2 * (linear_term_grad * M).sum()
        grad -= linear_term_grad
    
        loss += constant
        regularization_term = torch.sum(torch.Tensor([(regularizer[0] * X).sum() for regularizer, X in zip(regularization, Xs)]))
        loss += regularization_term
        if verbose:
            print(f"M regularization term: {regularization_term}")
            print(f"M constant term: {constant}")
            print(f"M constant magnitude: {constant_magnitude}")
        loss /= 2


        if M_constraint == 'simplex':
            grad.sub_(grad.sum(0, keepdim=True))

        return loss.item(), grad
    
    def estimate_M_nag(M, verbose=False):
        """Estimate M using Nesterov accelerated gradient descent.

        Args:
            M (torch.Tensor) : current estimate of meteagene parameters
        """
        loss, grad = compute_loss_and_gradient(M, verbose=verbose)
        if verbose:
            print(f"M NAG Initial Loss: {loss}")

        step_size = 1 / torch.linalg.eigvalsh(quadratic_factor).max().item()
        loss = np.inf
        
        optimizer = NesterovGD(M.clone(), step_size)
        for epoch in progress_bar:
            loss_prev = loss
            M_prev = M.clone()

            # Update M
            loss, grad = compute_loss_and_gradient(M)
            M = optimizer.step(grad)
            M = project_M(M, M_constraint)
            optimizer.set_parameters(M)

            dloss = loss_prev - loss
            dM = (M_prev - M).abs().max().item()
            stop_criterion = dM < tol and epoch > 5
            assert not np.isnan(loss)
            if epoch % 1000 == 0 or stop_criterion:
                progress_bar.set_description(
                    f'Updating M: loss = {loss:.1e}, '
                    f'%δloss = {dloss / loss:.1e}, '
                    f'δM = {dM:.1e}'
                    # f'lr={step_size_scale:.1e}'
                )
            if stop_criterion:
                break
        
        loss, grad = compute_loss_and_gradient(M, verbose=True)
        if verbose:
            print(f"M NAG Final Loss: {loss}")

        return M
    
    if backend_algorithm == 'mu':
        for epoch in progress_bar:
            loss = (((M @ quadratic_factor) * M).sum() - 2 * (M * linear_term).sum() + constant) / 2
            loss = loss.item()
            numerator = linear_term
            denominator = M @ quadratic_factor
            multiplicative_factor = numerator / denominator

            M_prev = M.clone()
            # multiplicative_factor.clip_(max=10)
            M *= multiplicative_factor
            M = project_M(M, M_constraint)
            dM = M_prev.sub(M).abs_().max().item()

            stop_criterion = dM < tol and epoch > 5
            if epoch % 1000 == 0 or stop_criterion:
                progress_bar.set_description(
                    f'Updating M: loss = {loss:.1e}, '
                    f'%δloss = {(loss_prev - loss) / loss:.1e}, '
                    f'δM = {dM:.1e}'
                )
            if stop_criterion:
                break

    elif backend_algorithm == 'gd':
        step_size = 1 / torch.linalg.eigvalsh(quadratic_factor).max().item()
        step_size_scale = 1
        loss, grad = compute_loss_and_gradient(M)
        dM = dloss = np.inf
        for epoch in progress_bar:
            M_new = M.sub(grad, alpha=step_size * step_size_scale)
            M_new = project_M(M_new, M_constraint)
            loss_new, grad_new = compute_loss_and_gradient(M_new)
            if loss_new < loss or step_size_scale == 1:
                dM = (M_new - M).abs().max().item()
                dloss = loss - loss_new
                M[:] = M_new
                loss = loss_new
                grad = grad_new
                step_size_scale *= 1.1
            else:
                step_size_scale *= .5
                step_size_scale = max(step_size_scale, 1.)

            stop_criterion = dM < tol and epoch > 5
            if epoch % 1000 == 0 or stop_criterion:
                progress_bar.set_description(
                    f'Updating M: loss = {loss:.1e}, '
                    f'%δloss = {dloss / loss:.1e}, '
                    f'δM = {dM:.1e}, '
                    f'lr={step_size_scale:.1e}'
                )
            if stop_criterion:
                break

    elif backend_algorithm == 'gd Nesterov':
        M = estimate_M_nag(M, verbose=verbose)

    else:
        raise NotImplementedError
    
    return M

def estimate_Sigma_x_inv(Xs, Sigma_x_inv, spatial_flags, lambda_Sigma_x_inv, betas, optimizer, context, datasets, Sigma_x_inv_bar=None, n_epochs=1000):
    """Optimize Sigma_x_inv parameters.

   
    Differential mode:
    grad = ( M XT X - YT X ) / ( σ_yx^2 ) + λ_Sigma_x_inv ( Sigma_x_inv - Sigma_x_inv_bar )

    Args:
        Xs: list of latent expression embeddings for each FOV.
        Sigma_x_inv: previous estimate of Σx-1

    """
    num_edges_per_fov = [sum(map(len, dataset.obs["adjacency_list"])) for dataset in datasets]

    if not any(sum(map(len, dataset.obs["adjacency_list"])) > 0 and u for dataset, u in zip(datasets, spatial_flags)):
        return

    linear_term_coefficient = torch.zeros_like(Sigma_x_inv).requires_grad_(False)
    # Zs = [torch.tensor(X / np.linalg.norm(X, axis=1, ord=1, keepdims=True), **context) for X in Xs]
    size_factors = [torch.linalg.norm(X, axis=1, ord=1, keepdim=True) for X in Xs ]
    Zs = [X / size_factor for X, size_factor in zip(Xs, size_factors)]
    nus = [] # sum of neighbors' z
    weighted_total_cells = 0
    for Z, dataset, use_spatial, beta in zip(Zs, datasets, spatial_flags, betas):
        adjacency_matrix = dataset.obsp["adjacency_matrix"]
        adjacency_list = dataset.obs["adjacency_list"]
        if use_spatial:
            nu = adjacency_matrix @ Z
            linear_term_coefficient = linear_term_coefficient.addmm_(Z.T, nu, alpha=beta)
        else:
            nu = None
        nus.append(nu)
        weighted_total_cells += beta * sum(map(len, adjacency_list))
        del Z, adjacency_matrix
    # linear_term_coefficient = (linear_term_coefficient + linear_term_coefficient.T) / 2 # should be unnecessary as long as adjacency_list is symmetric

    history = []
    Sigma_x_inv.requires_grad_(True)

    if optimizer:
        loss_prev, loss = np.inf, np.nan
        progress_bar = trange(n_epochs, desc='Updating Σx-1')
        Sigma_x_inv_best, loss_best, epoch_best = None, np.inf, -1
        dSigma_x_inv = np.inf
        early_stop_epoch_count = 0
        for epoch in progress_bar:
            optimizer.zero_grad()

            # Compute loss 
            linear_term = Sigma_x_inv.view(-1) @ linear_term_coefficient.view(-1)
            regularization = lambda_Sigma_x_inv * Sigma_x_inv.pow(2).sum() * weighted_total_cells / 2
            if Sigma_x_inv_bar is not None:
                regularization += lambda_Sigma_x_inv * 100 * (Sigma_x_inv_bar - Sigma_x_inv).pow(2).sum() * weighted_total_cells / 2
            
            log_partition_function = 0
            for Z, nu, beta in zip(Zs, nus, betas):
                if nu is None:
                    continue
                assert torch.isfinite(nu).all()
                assert torch.isfinite(Sigma_x_inv).all()
                eta = nu @ Sigma_x_inv
                logZ = integrate_of_exponential_over_simplex(eta)
                log_partition_function += beta * logZ.sum()

            loss = (linear_term + regularization + log_partition_function) / weighted_total_cells
            # if epoch == 0: print(loss.item())

            if loss < loss_best:
                Sigma_x_inv_best = Sigma_x_inv.clone().detach()
                loss_best = loss.item()
                epoch_best = epoch

            history.append((Sigma_x_inv.detach().cpu().numpy(), loss.item()))

            with torch.no_grad():
                Sigma_x_inv_prev = Sigma_x_inv.clone().detach()

            loss.backward()
            Sigma_x_inv.grad = (Sigma_x_inv.grad + Sigma_x_inv.grad.T) / 2
            optimizer.step()
            # with torch.no_grad():
            #   Sigma_x_inv -= Sigma_x_inv.mean()

            loss = loss.item()
            dloss = loss_prev - loss
            loss_prev = loss

            with torch.no_grad():
                dSigma_x_inv = Sigma_x_inv_prev.sub(Sigma_x_inv).abs().max().item()
            
            progress_bar.set_description(
                f'Updating Σx-1: loss = {dloss:.1e} -> {loss:.1e} '
                f'δΣx-1 = {dSigma_x_inv:.1e} '
                f'Σx-1 range = {Sigma_x_inv.min().item():.1e} ~ {Sigma_x_inv.max().item():.1e}'
            )

            if Sigma_x_inv.grad.abs().max() < 1e-4 and dSigma_x_inv < 1e-4:
                early_stop_epoch_count += 1
            else:
                early_stop_epoch_count = 0
            if early_stop_epoch_count >= 10 or epoch > epoch_best + 100:
                break

        Sigma_x_inv = Sigma_x_inv_best

        return Sigma_x_inv, loss * weighted_total_cells
    else:
        pass
    
    Sigma_x_inv.requires_grad_(False)



#!/usr/bin/env python
import random
import numpy as np

import scipy
from scipy.spatial import Delaunay
from scipy.spatial.distance import pdist, cdist, squareform
from scipy.stats import gaussian_kde, gamma, truncnorm, truncexpon, expon, bernoulli, dirichlet

import emcee
from numpy.linalg import inv
from numpy.random import multivariate_normal

from sklearn.decomposition import NMF

import umap
import pickle as pkl
import seaborn as sns
import pandas as pd
import networkx as nx
import sys
import os
from pathlib import Path

import matplotlib.pyplot as plt
from matplotlib.colors import hsv_to_rgb
from matplotlib.colors import ListedColormap
        
import seaborn as sns

def sample_gaussian(sigma, m, N=1):
    """

    TODO: Not exactly sure how this works.
    
    """
    K = len(sigma)
    assert sigma.shape[0] == sigma.shape[1]
    assert len(m) == K
    
    # Box-Muller Method
    L = np.linalg.cholesky(sigma)
    n_z = K + (K % 2)
    x = np.zeros((n_z, N))
    num_samples = 0
    while True:
        n_valid = 0
        while True:
            z = 2*np.random.rand(2) - 1
            if (z[0]**2 + z[1]**2 <= 1):
                r = np.linalg.norm(z)
                x[n_valid, num_samples] = z[0]*np.sqrt(-2*np.log(r**2)/r**2)
                x[n_valid + 1, num_samples] = z[1]*np.sqrt(-2*np.log(r**2)/r**2)
                n_valid += 2
            if n_valid == n_z:
                num_samples += 1
                break
        if num_samples == N:
            break
            
    # if K is odd, there will be one extra sample, so throw it away
    x = x[0:K, :]
    x = np.dot(L, x) + np.expand_dims(m, -1)
    
    return np.squeeze(x)

# def sample_truncated_gaussian(covariance_matrix, means, num_walkers, num_steps=10000):
#     position = multivariate_normal(means, covariance_matrix, size=num_walkers)
    
    
    
def sample_truncated_gaussian(covariance_matrix, means, num_walkers, num_steps=10000):
    """Sample gene expression vectors from a multivariate truncated Gaussian distribution.
    
    See https://stackoverflow.com/a/20188431/13952002 for details
    
    Args:
        covariance_matrix:
        means:
        num_walkers:
    Return:
    
    Todo:
        Finish docstring
    """
    num_genes = len(covariance_matrix)
    bounds = np.tile([0, np.inf], (1, num_genes))
    
    def lnprob_trunc_norm(x, means, bounds, covariance_matrix):
        if np.any(x < bounds[:,0]) or np.any(x > bounds[:,1]):
            return -np.inf
        else:
            return -0.5*(x-means).dot(inv(covariance_matrix)).dot(x-means)
    
    sampler = emcee.EnsembleSampler(num_walkers, num_genes, lnprob_trunc_norm, args = (means, bounds, covariance_matrix), moves=[(emcee.moves.StretchMove(live_dangerously=True), 1)])
    position = multivariate_normal(means, covariance_matrix, size=num_walkers)

#     plt.figure()
#     plt.scatter(position[:, 0], position[:, 1])
    
    position, prob, state = sampler.run_mcmc(position, num_steps, skip_initial_state_check=True)
   
    # Some data may still be negative if the number of steps in the simulation is too low;
    # in this case, simply make them positive
    position = np.abs(position)
#     plt.scatter(position[:, 0], position[:, 1])
#     plt.show()
    
    return position
    
def sample_2D_points(num_points, minimum_distance):
    """Generate 2D samples that are at least minimum_distance apart from each other.
    
    """
    # TODO: Implement Poisson disc sampling for a vectorized operation
    
    points = np.zeros((num_points, 2))
    points[0] = np.random.random_sample(2)
    for index in range(1, num_points):
        while True:
            point = np.random.random_sample((1, 2))
            distances = cdist(points[:index], point)
            if np.min(distances) > minimum_distance:
                points[index] = point
                break
                
    return points


def generate_affinity_matrix(points, tau=1.0, method="delaunay"):
    """Given a set of 2D spatial coordinates, generate an affinity matrix.

    Can optionally use Delaunay triangulation or simple distance thresholding.
    """

    num_cells = len(points)
    if method == "delaunay":
        affinity_matrix = np.zeros((num_cells, num_cells))
        distance_matrix = np.zeros((num_cells, num_cells))
        triangulation = Delaunay(points)
        for triangle in triangulation.simplices:
            affinity_matrix[triangle[0], triangle[1]] = 1
            distance_matrix[triangle[0], triangle[1]] = np.linalg.norm(points[triangle[0]] - points[triangle[1]])
            affinity_matrix[triangle[1], triangle[2]] = 1
            distance_matrix[triangle[1], triangle[2]] = np.linalg.norm(points[triangle[1]] - points[triangle[2]])
            affinity_matrix[triangle[2], triangle[0]] = 1
            distance_matrix[triangle[2], triangle[0]] = np.linalg.norm(points[triangle[2]] - points[triangle[0]])
                  
        threshold = np.percentile(distance_matrix[np.nonzero(distance_matrix)], 95)
        affinity_matrix[distance_matrix > threshold] = 0

    else:
        disjoint_nodes = True
        while(disjoint_nodes):
            N = points.shape[0]
            # Construct graph
            distances = squareform(pdist(p))
            affinity_matrix = distances < tau
            identity_matrix = np.identity(N, dtype='bool')
            affinity_matrix = affinity_matrix * ~identity_matrix
            graph = nx.from_numpy_matrix(affinity_matrix)
            if not nx.is_connected(graph):
                # increase tau by 10% and repeat
                tau = 1.1*tau
                print('Graph is not connected, increasing tau to %s', tau)
            else:
                disjoint_nodes = False
                
    return affinity_matrix


def synthesize_metagenes(num_genes, num_real_metagenes, n_noise_metagenes, real_metagene_parameter, noise_metagene_parameter,  metagene_variation_probabilities, original_metagenes=None, replicate_variability=None, normalize=True):
    """Synthesize related metagenes according to the metagene_variation_probabilities vector.
    
    Creates num_real_metagenes synthetic metagenes using a random Gamma distribution with
    shape parameter real_metagene_parameter. For each metagene i, if dropout_probabilities[i] != 0,
    randomly permutes a metagene_variation_probabilities[i] fraction of metagene i-1 to create metagene i;
    otherwise, creates a new random metagene. In addition, adds n_noise_metagenes parameterized by
    a Gamma distribution with shape parameter noise_metagene_parameter.
    """
    
    num_metagenes = num_real_metagenes + n_noise_metagenes 
    metagenes = np.zeros((num_metagenes, num_genes))
    
    for index in range(num_real_metagenes):
        variation_probability = metagene_variation_probabilities[index]

        if variation_probability == 0 and not replicate_variability:
            metagene = gamma.rvs(real_metagene_parameter, size=num_genes)
            metagenes[index] = metagene
        else:
            if variation_probability == 0:
                metagene = original_metagenes[index].copy()
                variation_probability = replicate_variability
                
            mask = bernoulli.rvs(variation_probability, size=num_genes).astype('bool')
            perturbed_metagene = metagene

            perturbations = gamma.rvs(real_metagene_parameter, size=np.sum(mask))
            if original_metagenes is not None:
                # Use dirichlet distribution
                perturbations *= np.sum(metagene[mask]) / np.sum(perturbations)
            perturbed_metagene[mask] = perturbations
        
            metagenes[index] = perturbed_metagene
            
    for index in range(num_real_metagenes, num_metagenes):
        metagenes[index] = gamma.rvs(noise_metagene_parameter, size=num_genes)
        
    metagenes = metagenes
    
    if normalize:
        metagenes = metagenes / np.sum(metagenes, axis=1, keepdims=True)
       
    return metagenes


def synthesize_cell_embeddings(points, distributions, cell_type_definitions, mask_conditions, num_cells, n_noise_metagenes=3, signal_sigma_x=0.1, background_sigma_x=0.03, sigma_x_scale=1.0):
    """Generate synthetic cell embeddings.
    
    """
    
    num_patterns, num_cell_types, num_real_metagenes = cell_type_definitions.shape
    num_metagenes = num_real_metagenes + n_noise_metagenes
    
    sigma_x = np.concatenate([np.full(num_real_metagenes, signal_sigma_x), np.full(n_noise_metagenes, background_sigma_x)])
    sigma_x = sigma_x * sigma_x_scale

    cell_type = np.zeros((num_cells), dtype='int')
    Z = np.zeros((num_cells, num_metagenes))
    X = np.zeros((num_cells, num_metagenes))
    
    for pattern_index in range(len(mask_conditions)):
        pattern = mask_conditions[pattern_index](points)
        cell_type_definition = cell_type_definitions[pattern_index]
        distribution = distributions[pattern_index]
        
        cell_indices, = np.nonzero(pattern)
        random.shuffle(cell_indices)
        partition_indices = (np.cumsum(distribution) * len(cell_indices)).astype(int)
        partitions = np.split(cell_indices, partition_indices[:-1])
        
        for cell_type_index in range(len(cell_type_definition)):
            cell_type_composition = cell_type_definition[cell_type_index]
            partition = partitions[cell_type_index]
            cell_type[partition] = cell_type_index
            Z[partition, :num_real_metagenes] = cell_type_composition
        
   
    # Extrinsic factors
    Z[:, num_real_metagenes:num_metagenes] = 0.05

    # TODO: vectorize
    for cell in range(num_cells):
        for metagene in range(num_metagenes):
            X[cell, metagene] = sigma_x[metagene]*truncnorm.rvs(-Z[cell, metagene]/sigma_x[metagene], 100) + Z[cell, metagene]
    X = X * (Z > 0)
    X = X.T
    X = X / np.sum(X, axis=0)
    
    return X, cell_type

def perturb_genes(gene_expression, num_metagenes, first_threshold=.2, second_threshold=.2, shape=2.5):
    """Randomly perturb gene expression values.

    """
    genes, num_samples = gene_expression.shape
    random_sample_size = int(first_threshold * genes)
    for sample in range(num_samples):
        random_indices = random.sample(range(genes), random_sample_size)
        gene_expression[random_indices, sample] = gamma.rvs(shape, size=random_sample_size) / float(genes)
    random_sample_size = int(second_threshold * genes)
    indices = random.sample(range(genes), random_sample_size)
    gene_expression[indices, :] = (gamma.rvs(shape, size=(random_sample_size*num_samples)) / float(genes)).reshape((random_sample_size, num_samples))

    return gene_expression

class SyntheticDataset:
    """Synthetic mouse brain cortex dataset.
    
    This class provides methods for initializing a semi-random mouse cortex spatial
    transcriptomics dataset, as well as methods to visualize aspects of the dataset.
    
    """
    
    def __init__(self, distributions, cell_type_definitions, mask_conditions, metagene_variation_probabilities,
                 parameters, parent_directory, shared_metagenes=None, key=''):
        self.num_metagenes = parameters["num_real_metagenes"] + parameters['num_noise_metagenes']
        self.num_cells = parameters['num_cells']
        self.num_genes = parameters["num_genes"]
        
        # TODO: make color work for variable number of colors
        self.colors = {0: 'darkkhaki', 1: 'mediumspringgreen', 2: 'greenyellow', 3: '#95bfa6',
                       4: 'violet', 5: 'firebrick',
                       6: 'deepskyblue', 7: 'darkslateblue'}
        
        print('Synthesizing X, A, and p...')
        self.num_replicates = parameters['num_replicates']
        self.sig_y = parameters['sigY_scale']
        if isinstance(self.sig_y, float):
            self.sig_y *= np.identity(self.num_genes) / self.num_genes
            self.variance_y = (self.sig_y**2)
        elif isinstance(self.sig_y, dict):
            self.sig_y = {cell_type: cell_specific_sig_y / self.num_genes for cell_type, cell_specific_sig_y in self.sig_y.items()}
            random_key = next(iter(self.sig_y))
            if isinstance(self.sig_y[random_key], float):
                self.sig_y = {cell_type: cell_specific_sig_y * np.identity(self.num_genes) for cell_type, cell_specific_sig_y in self.sig_y.items()}
            
            self.variance_y = {cell_type: cell_specific_sig_y ** 2 for cell_type, cell_specific_sig_y in self.sig_y.items()}
        
        # self.sig_y = float(parameters['sigY_scale']) / self.num_genes
        self.metagenes = np.zeros((self.num_replicates, self.num_genes, self.num_metagenes))
        self.Y = np.zeros((self.num_replicates, self.num_genes, self.num_cells))
        self.cell_embeddings = np.zeros((self.num_replicates, self.num_metagenes, self.num_cells))
        self.points = np.zeros((self.num_replicates, self.num_cells, 2))
        self.affinity_matrices = np.zeros((self.num_replicates, self.num_cells, self.num_cells))
        self.cell_types = np.zeros((self.num_replicates, self.num_cells))
        
        minimum_distance = 0.75 / np.sqrt(self.num_cells)
        tau = minimum_distance * 2.2
        for replicate in range(self.num_replicates):
            print('Synthesizing M...')
            if replicate == 0:
                self.metagenes[replicate] = synthesize_metagenes(self.num_genes, parameters["num_real_metagenes"],
                                                  parameters['num_noise_metagenes'],
                                                  parameters["real_metagene_parameter"], parameters["noise_metagene_parameter"],
                                                  metagene_variation_probabilities=metagene_variation_probabilities)
            else:
                self.metagenes[replicate] = synthesize_metagenes(self.num_genes, parameters["num_real_metagenes"],
                                                  parameters['num_noise_metagenes'],
                                                  parameters["real_metagene_parameter"], parameters["noise_metagene_parameter"],
                                                  metagene_variation_probabilities=metagene_variation_probabilities,
                                                  original_metagenes=self.metagenes[replicate-1].T,
                                                  replicate_variability=parameters["replicate_variability"])
        
            p_i = sample_2D_points(self.num_cells, minimum_distance)
            A_i = generate_affinity_matrix(p_i, tau)
            X_i, C_i = synthesize_cell_embeddings(p_i, distributions, cell_type_definitions, mask_conditions, self.num_cells, n_noise_metagenes=parameters["num_noise_metagenes"])

            self.S = gamma.rvs(self.num_metagenes, scale=parameters['lambda_s'], size=self.num_cells)
            self.affinity_matrices[replicate] = A_i
            self.points[replicate] = p_i
            self.cell_embeddings[replicate] = X_i * self.S
            self.cell_types[replicate] = C_i

        print(self.metagenes.sum(axis=1))
        print('Synthesizing Y...')
        for replicate in range(self.num_replicates):
            Y_i = np.matmul(self.metagenes[replicate], self.cell_embeddings[replicate])
            self.Y[replicate] = Y_i
#             variance_y = (self.sig_y**2) * np.identity(self.num_genes)
            
            for cell, cell_type in zip(range(self.num_cells), self.cell_types[replicate]):
                if isinstance(self.variance_y, dict):
                    cell_type_variance_y = self.variance_y[int(cell_type)]
                else:
                    cell_type_variance_y = self.variance_y
                    
                self.Y[replicate][:, cell] = np.abs(sample_gaussian(cell_type_variance_y, Y_i[:, cell]))
                
            if parameters['gene_replace_prob'] > 0 and parameters['element_replace_prob'] > 0:
                self.Y[replicate] = perturb_genes(Y_i, self.num_metagenes, first_threshold=parameters['gene_replace_prob'],
                                        second_threshold=parameters['element_replace_prob'],
                                        shape=self.num_metagenes)
                
        # gene_ind variable is just all genes -- we don't remove any
        # TODO: remove this field
        self.gene_ind = range(self.num_genes)
        
        # create empty Sig_x_inverse, since it is not used in this data generation
        self.sigma_x_inverse = np.zeros((self.num_metagenes, self.num_metagenes))

#         data_subdirectory = 'synthetic_{}_{}_{}_{:.0f}_{}'
#         data_subdirectory = data_subdirectory.format(self.num_cells, self.num_genes, "covariance",
#                                                          parameters['gene_replace_prob']*100, key)
#         
#         self.data_directory = Path(parent_directory) / data_subdirectory
#         self.initialize_data_directory()
        
        print('Finished')

    def initialize_data_directory(self):
        """Initialize data directory structure on user file system.
        
        """
        (self.data_directory / "files").mkdir(parents=True, exist_ok=True)
        (self.data_directory / "logs").mkdir(parents=True, exist_ok=True)
        (self.data_directory / "scripts").mkdir(parents=True, exist_ok=True)
        (self.data_directory / "plots").mkdir(parents=True, exist_ok=True)

    def plot_cells_UMAP(self, replicate=0, latent_space=False, cell_types=None,colors=None,save_figure=False, normalize=True, annotate=False):
        """Plot synthesized cells using UMAP.
        """
        # TODO: fix colors to be more compatible with variable cell types..
        
        if latent_space:
            gene_expression = self.cell_embeddings[replicate].T
        else:
            gene_expression = self.Y[replicate].T
            
        num_cells, num_features = gene_expression.shape
        C_i = self.cell_types[replicate]
        
        if not colors:
            colors = {0: 'darkkhaki', 1: 'mediumspringgreen', 2: 'greenyellow', 3: '#95bfa6',
                      4: 'violet', 5: 'firebrick', 6: 'gold',
                      7: 'deepskyblue', 8: 'darkslateblue', 9: 'gainsboro'}
            
       
        cell_type_names, cell_type_index = np.unique(C_i, return_index=True)
        sort_index = np.argsort(cell_type_names)
        unique_cell_types, cell_type_index = cell_type_names[sort_index], cell_type_index[sort_index]
        palette = sns.color_palette("husl", len(unique_cell_types))
        sns.set_palette(palette)
        
        colormap = ListedColormap(palette)

        if normalize:
            gene_expression = (gene_expression - np.average(gene_expression, axis=0))
            gene_expression_std = gene_expression.std(axis=0)
            for feature in range(num_features):
                if gene_expression_std[feature] != 0:
                    gene_expression[:, feature] = np.divide(gene_expression[:, feature], gene_expression_std[feature])

        # TODO: cleanup unnecessary lines
        gene_expression_reduced = umap.UMAP(
                        n_components=2,
                        #         spread=1,
                        n_neighbors=10,
                        min_dist=0.3,
                        #         learning_rate=100,
                        #         metric='euclidean',
                        #         metric='manhattan',
                        #         metric='canberra',
                        #         metric='braycurtis',
                        #         metric='mahalanobis',
                        #         metric='cosine',
                        #         metric='correlation',
                        ).fit_transform(gene_expression)

        fig = plt.figure(figsize=(7, 5))
        ax = fig.add_axes([0.1, 0.1, .8, .8])
        for color, cell_type in np.ndenumerate(unique_cell_types):
            index = (C_i == cell_type)
            ax.scatter(gene_expression_reduced[index, 0], gene_expression_reduced[index, 1], alpha=.7, c=colormap(color), label=cell_type)
        ax.set_xticks([])
        ax.set_yticks([])
        ax.yaxis.set_label_position("right")
        plt.legend(bbox_to_anchor=(1.04,0), loc="lower left", borderaxespad=0)
        
        if annotate:
            for label, index in zip(unique_cell_types, cell_type_index):
                ax.annotate(label, (gene_expression_reduced[index, 0], gene_expression_reduced[index, 1]))

        plt.show()
        
        if save_figure:
            plt.savefig(self.data_directory / "plots" / 'synthesized_data_umap.png')

    def plot_metagenes(self, order_genes=True):
        """Plot density map of metagenes in terms of constituent genes.
        
        """
        
        if order_genes:
            mclust = scipy.cluster.hierarchy.linkage(self.metagenes, 'ward')
            mdendro = scipy.cluster.hierarchy.dendrogram(mclust, no_plot=True)
            plt.imshow(self.metagenes[mdendro['leaves']], aspect='auto')
        else:
            plt.imshow(self.metagenes, aspect='auto')
            
        plt.xlabel('Metagene ID')
        plt.ylabel('Gene ID')
        plt.show()

    def plot_cell_types(self, replicate=0, save_figure=False, colors=None):
        """Plot cells in situ using cell type labels.
        
        """
        
        points = self.points[replicate]
        affinity_matrix = self.affinity_matrices[replicate]
        cell_types = self.cell_types[replicate]
        if not colors:
            colors = {0: 'sandybrown', 1: 'lightskyblue',
                      2: 'mediumspringgreen', 3: 'palegreen',
                      4: 'greenyellow', 5: 'darkseagreen',
                      6: 'burlywood', 7: 'orangered', 8: 'firebrick',
                      9: 'gold', 10: 'mediumorchid', 11: 'magenta',
                      12: 'palegoldenrod', 13: 'gainsboro', 14: 'teal',
                      15: 'darkslateblue'}
        df = pd.DataFrame({'X': points[:, 0], 'Y': points[:, 1], 'cell_type': cell_types})
        fig = plt.figure(figsize=(12, 6))
        ax = fig.add_axes([0, 0, 1, 1])
        
        for (source, destination) in zip(*np.where(affinity_matrix == 1)):
            plt.plot([points[source, 0], points[destination, 0]],
                [points[source, 1], points[destination, 1]], color="gray", linewidth=1)
        
        sns.scatterplot(data=df, x='X', y='Y', hue='cell_type', ax=ax, palette=colors,
                        legend=False, hue_order=list(set(cell_types)), size_norm=10.0)
        plt.show()

    def plot_metagenes_in_situ(self, replicate=0, save_figure=False):
        """Plot metagene values per cell in-situ.
        
        """
        
        points = self.points[replicate]
        cell_embeddings = self.cell_embeddings[replicate]
        for metagene in range(self.num_metagenes):
            fig = plt.figure(figsize=(5, 3))
            ax = fig.add_axes([.1, .1, .8, .8])
            ax.set_ylabel('Metagene {}'.format(metagene))
            sca = ax.scatter(points[:, 0], points[:, 1], c=cell_embeddings[metagene], s=23, cmap=plt.get_cmap('Blues'), vmin=0)
            fig.colorbar(sca)
            ax.set_yticks([])
            ax.set_xticks([])
            plt.show()
            if save_figure:
                plt.savefig(self.data_directory / "plots" / ('synthetic_metagene_{}.png'.format(metagene)))

    def plot_hidden_states(self, replicate=0, save_figure=False):
        """Plot synthetic cell embeddings.
        
        """
        cell_embeddings = self.cell_embeddings[replicate]
        image = plt.imshow(cell_embeddings, aspect='auto')
        plt.xlabel('Cell ID')
        plt.ylabel('Metagene ID')
        plt.colorbar(image)
        plt.show()
